/* -*- Mode: C; c-file-style: "stroustrup" -*- */
#include "gsttextoverlay.h"


static GstElementDetails textoverlay_details = {
    "Text Overlay", 
    "Filter/Video", 
    "LGPL", 
    "Adds text strings on top of a video buffer", 
    VERSION,
    "Gustavo J. A. M. Carneiro <gjc@inescporto.pt>", 
    "(C) 2003", 
};



enum {
    ARG_0, 
    ARG_TEXT, 
};

GST_PAD_TEMPLATE_FACTORY(textoverlay_src_template_factory, 
			 "src", 
			 GST_PAD_SRC, 
			 GST_PAD_ALWAYS, 
			 GST_CAPS_NEW(
			     "textoverlay_src", 
			     "video/raw", 
			     "format",		GST_PROPS_LIST(
				 GST_PROPS_FOURCC(GST_STR_FOURCC("I420"))
				 ), 
			     "width", 		GST_PROPS_INT_RANGE(0, G_MAXINT),
			     "height",		GST_PROPS_INT_RANGE(0, G_MAXINT)
			     )
    )

GST_PAD_TEMPLATE_FACTORY(textoverlay_sink_template_factory, 
			 "sink", 
			 GST_PAD_SINK, 
			 GST_PAD_ALWAYS, 
			 GST_CAPS_NEW(
			     "textoverlay_sink", 
			     "video/raw", 
			     "format",		GST_PROPS_LIST(
				 GST_PROPS_FOURCC(GST_STR_FOURCC("I420"))
				 ), 
			     "width", 		GST_PROPS_INT_RANGE(0, G_MAXINT), 
			     "height",		GST_PROPS_INT_RANGE(0, G_MAXINT)
			     )
    )
    
static void                  gst_textoverlay_class_init   (GstTextOverlayClass *klass);
static void                  gst_textoverlay_init         (GstTextOverlay      *overlay);
static void                  gst_textoverlay_set_property (GObject             *object,
							   guint                prop_id,
							   const GValue        *value,
							   GParamSpec          *pspec);
static void                  gst_textoverlay_get_property (GObject             *object,
							   guint                prop_id,
							   GValue              *value,
							   GParamSpec          *pspec);
static void                  gst_textoverlay_chain        (GstPad              *pad,
							   GstBuffer           *buf);
static GstElementStateReturn gst_textoverlay_change_state (GstElement          *element);
static void                  gst_textoverlay_finalize     (GObject             *object);


static GstElementClass *parent_class = NULL;
/*static guint gst_textoverlay_signals[LAST_SIGNAL] = { 0 }; */


GType
gst_textoverlay_get_type(void)
{
    static GType textoverlay_type = 0;

    if (!textoverlay_type) {
	static const GTypeInfo textoverlay_info = {
	    sizeof(GstTextOverlayClass),
	    NULL, 
	    NULL, 
	    (GClassInitFunc)gst_textoverlay_class_init, 
	    NULL, 
	    NULL, 
	    sizeof(GstTextOverlay), 
	    0, 
	    (GInstanceInitFunc)gst_textoverlay_init, 
	};
	textoverlay_type = g_type_register_static(GST_TYPE_ELEMENT, "GstTextOverlay",
						  &textoverlay_info, 0);
    }
    return textoverlay_type;
}


static void
gst_textoverlay_class_init(GstTextOverlayClass *klass)
{
    GObjectClass *gobject_class;
    GstElementClass *gstelement_class;

    gobject_class = (GObjectClass*)klass;
    gstelement_class = (GstElementClass*)klass;

    parent_class = g_type_class_peek_parent(klass);

    gobject_class->finalize = gst_textoverlay_finalize;
    gobject_class->set_property = gst_textoverlay_set_property;
    gobject_class->get_property = gst_textoverlay_get_property;

    gstelement_class->change_state = gst_textoverlay_change_state;
    /* FIXME: make dpi configurable (property) */
    klass->pango_context = pango_ft2_get_context(72 /*dpi_x*/,
						 72 /*dpi_y*/);
    g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_TEXT,
				    g_param_spec_string("text","text",
							"Text to be display,"
							" in pango markup format.",
							"", G_PARAM_WRITABLE));
}


static void
resize_bitmap(FT_Bitmap *bitmap, int width, int height)
{
    int pitch = (width|3) + 1, size = pitch*height;

    /* no need to keep reallocating; just keep the maximum size so far */
    if (width < bitmap->width && height < bitmap->rows) {
	memset(bitmap->buffer, 0, bitmap->rows*bitmap->pitch);
	return;
    }
    if (!bitmap->buffer) {
	/* initialize */
	bitmap->pixel_mode = FT_PIXEL_MODE_GRAY;
	bitmap->num_grays  = 256;
    }
    if (bitmap->buffer)
	bitmap->buffer = g_realloc (bitmap->buffer, size);
    else
	bitmap->buffer = g_malloc (size);
    bitmap->rows  = height;
    bitmap->width = width;
    bitmap->pitch = pitch;
    memset(bitmap->buffer, 0, size);
}

static void
set_text(GstTextOverlay *overlay, const gchar *text)
{
    PangoRectangle ink_rect, logical_rect;
    pango_layout_set_markup(overlay->layout, text, -1);
    pango_layout_get_pixel_extents(overlay->layout, &ink_rect, &logical_rect);
    resize_bitmap(&overlay->bitmap, ink_rect.width, ink_rect.height + ink_rect.y);
    pango_ft2_render_layout(&overlay->bitmap, overlay->layout, 0, 0);
}

static GstPadLinkReturn
gst_textoverlay_sinkconnect (GstPad *pad, GstCaps *caps)
{
    GstTextOverlay *overlay;

    overlay = GST_TEXTOVERLAY(gst_pad_get_parent (pad));
  
    if (!GST_CAPS_IS_FIXED (caps))
	return GST_PAD_LINK_DELAYED;

    overlay->width = overlay->height = 0;
    gst_caps_get_int(caps, "width", &overlay->width);
    gst_caps_get_int(caps, "height", &overlay->height);

    return gst_pad_try_set_caps(overlay->srcpad, caps);
}

static void
gst_text_overlay_blit_yuv420(GstTextOverlay *overlay, FT_Bitmap *bitmap, 
			     guchar *pixbuf, int x0, int y0)
{
    register int x, y;
#define yuv420_pixel(x, y) (*(pixbuf + overlay->width*(y) + (x)))

    /* FIXME: the text is not getting drawn at the correct position;
     * my knowledge about YUV format sucks, I need help. */

    /* paint black margin, brute force approach */
    for (y = 0; y < bitmap->rows; y++) 
    {
	for (x = 0; x < bitmap->width; x++)
	{
	    register guint16 v;
	    int x1 = x0 + x;
	    int y1 = y0 + y;
	    if (x1 >= 0 && x1 < overlay->width && y1 >= 0 && y1 < overlay->height) {
		v = bitmap->buffer[y * bitmap->pitch + x];
		if (v) {
		    yuv420_pixel(x1 + 1, y1 + 1) = 0;
		    yuv420_pixel(x1 + 1, y1 - 1) = 0;
		    yuv420_pixel(x1 - 1, y1 + 1) = 0;
		    yuv420_pixel(x1 - 1, y1 - 1) = 0;
		}
	    }
	}
    }

    /* paint text */
    for (y = 0; y < bitmap->rows; y++) 
    {
	for (x = 0; x < bitmap->width; x++)
	{
	    register guint16 v;
	    int x1 = x0 + x;
	    int y1 = y0 + y;
	    if (x1 >= 0 && x1 < overlay->width && y1 >= 0 && y1 < overlay->height) {
		v = bitmap->buffer[y * bitmap->pitch + x];
		if (v)
		    yuv420_pixel(x1, y1) = v;
	    }
	}
    }

#undef yuv420_pixel
}

static void
gst_textoverlay_chain(GstPad *pad, GstBuffer *buf)
{
    GstTextOverlay *overlay;
    guchar         *pixbuf;

    g_return_if_fail(pad != NULL);
    g_return_if_fail(GST_IS_PAD(pad));
    g_return_if_fail(buf != NULL);
    overlay = GST_TEXTOVERLAY(gst_pad_get_parent(pad));
    g_return_if_fail(overlay != NULL);
    g_return_if_fail(GST_IS_TEXTOVERLAY(overlay));
    
    pixbuf = GST_BUFFER_DATA(buf);

    if (overlay->bitmap.buffer)
	gst_text_overlay_blit_yuv420(overlay, &overlay->bitmap, pixbuf,
				     /* vertical alignment: bottom */
				     overlay->height - overlay->bitmap.rows,
				     /* horizontal alignment: center */
				     overlay->width/2 - overlay->bitmap.width/2);
    gst_pad_push(overlay->srcpad, buf);
}


static GstElementStateReturn
gst_textoverlay_change_state(GstElement *element)
{
    GstTextOverlay *overlay;

    overlay = GST_TEXTOVERLAY(element);

    switch (GST_STATE_TRANSITION(element))
    {
    case GST_STATE_PAUSED_TO_PLAYING:
	break;
    case GST_STATE_PLAYING_TO_PAUSED:
	break;
    case GST_STATE_PAUSED_TO_READY:
	break;
    }

    parent_class->change_state(element);

    return GST_STATE_SUCCESS;
}

static void
gst_textoverlay_finalize(GObject *object)
{
    GstTextOverlay *overlay = GST_TEXTOVERLAY(object);
    if (overlay->layout) {
	g_object_unref(overlay->layout);
	overlay->layout = NULL;
    }
    if (overlay->bitmap.buffer) {
	g_free(overlay->bitmap.buffer);
	overlay->bitmap.buffer = NULL;
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
gst_textoverlay_init(GstTextOverlay *overlay)
{
    overlay->sinkpad = gst_pad_new_from_template(
	GST_PAD_TEMPLATE_GET(textoverlay_sink_template_factory), "sink");
    gst_pad_set_chain_function(overlay->sinkpad, gst_textoverlay_chain);
    gst_pad_set_link_function(overlay->sinkpad, gst_textoverlay_sinkconnect);
    gst_element_add_pad(GST_ELEMENT(overlay), overlay->sinkpad);

    overlay->srcpad = gst_pad_new_from_template(
	GST_PAD_TEMPLATE_GET(textoverlay_src_template_factory), "src");
    gst_element_add_pad(GST_ELEMENT(overlay), overlay->srcpad);

    overlay->layout = pango_layout_new(GST_TEXTOVERLAY_GET_CLASS(overlay)->pango_context);
    memset(&overlay->bitmap, 0, sizeof(overlay->bitmap));
    set_text(overlay, "<span size=\"xx-large\" weight=\"bold\">Hello World!</span>");
}


static void
gst_textoverlay_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
    GstTextOverlay *overlay;

    /* it's not null if we got it, but it might not be ours */
    g_return_if_fail(GST_IS_TEXTOVERLAY(object));
    overlay = GST_TEXTOVERLAY(object);

    switch (prop_id) {
    case ARG_TEXT:
	set_text(overlay, g_value_get_string(value));
	break;
    default:
	break;
    }
}

static void
gst_textoverlay_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
    GstTextOverlay *overlay;

    /* it's not null if we got it, but it might not be ours */
    g_return_if_fail(GST_IS_TEXTOVERLAY(object));
    overlay= GST_TEXTOVERLAY(object);

    switch (prop_id) {
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
	break;
    }
}

static gboolean
plugin_init(GModule *module, GstPlugin *plugin)
{
    GstElementFactory *factory;

    factory = gst_element_factory_new("textoverlay", GST_TYPE_TEXTOVERLAY, 
				      &textoverlay_details);
    g_return_val_if_fail(factory != NULL, FALSE);
    gst_element_factory_set_rank(factory, GST_ELEMENT_RANK_PRIMARY);

    gst_element_factory_add_pad_template(factory,
					 GST_PAD_TEMPLATE_GET(textoverlay_src_template_factory));
    gst_element_factory_add_pad_template(factory,
					 GST_PAD_TEMPLATE_GET(textoverlay_sink_template_factory));

    gst_plugin_add_feature(plugin, GST_PLUGIN_FEATURE(factory));

    return TRUE;
}

GstPluginDesc plugin_desc = {
    GST_VERSION_MAJOR, 
    GST_VERSION_MINOR, 
    "textoverlay", 
    plugin_init
};

